
# This file was generated by '_versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-04-20T13:28:42-0600",
 "dirty": false,
 "error": null,
 "full-revisionid": "3d0ee8b2dfbb825c532447c851172b491c0fab1a",
 "version": "1.2.6"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
